@@ -1,4 +1,4 @@
-# Copyright 2012-2015, Andrey Kislyuk and argcomplete contributors.
+# Copyright 2012-2019, Andrey Kislyuk and argcomplete contributors.
 # Licensed under the Apache License. See https://github.com/kislyuk/argcomplete for more info.
 
 from __future__ import absolute_import, division, print_function, unicode_literals
@@ -8,6 +8,7 @@ from . import completers, my_shlex as shlex
 from .compat import USING_PYTHON2, str, sys_encoding, ensure_str, ensure_bytes
 from .completers import FilesCompleter, SuppressCompleter
 from .my_argparse import IntrospectiveArgumentParser, action_is_satisfied, action_is_open, action_is_greedy
+from .shellintegration import shellcode
 
 _DEBUG = "_ARC_DEBUG" in os.environ
 
@@ -15,6 +16,11 @@ debug_stream = sys.stderr
 
 def debug(*args):
     if _DEBUG:
+        if USING_PYTHON2:
+            # debug_stream has to be binary mode in Python 2.
+            # Attempting to write unicode directly uses the default ascii conversion.
+            # Convert any unicode to bytes, leaving non-string input alone.
+            args = [ensure_bytes(x) if isinstance(x, str) else x for x in args]
         print(file=debug_stream, *args)
 
 BASH_FILE_COMPLETION_FALLBACK = 79
@@ -94,7 +100,9 @@ def split_line(line, point=None):
             if lexer.instream.tell() >= point:
                 return split_word(lexer.token)
             else:
-                raise ArgcompleteException("Unexpected internal state. Please report this bug at https://github.com/kislyuk/argcomplete/issues.")  # noqa
+                msg = ("Unexpected internal state. "
+                       "Please report this bug at https://github.com/kislyuk/argcomplete/issues.")
+                raise ArgcompleteException(msg)
 
 def default_validator(completion, prefix):
     return completion.startswith(prefix)
@@ -194,12 +202,6 @@ class CompletionFinder(object):
         comp_line = os.environ["COMP_LINE"]
         comp_point = int(os.environ["COMP_POINT"])
 
-        # Adjust comp_point for wide chars
-        if USING_PYTHON2:
-            comp_point = len(comp_line[:comp_point].decode(sys_encoding))
-        else:
-            comp_point = len(comp_line.encode(sys_encoding)[:comp_point].decode(sys_encoding))
-
         comp_line = ensure_str(comp_line)
         cword_prequote, cword_prefix, cword_suffix, comp_words, last_wordbreak_pos = split_line(comp_line, comp_point)
 
@@ -211,8 +213,15 @@ class CompletionFinder(object):
         start = int(os.environ["_ARGCOMPLETE"]) - 1
         comp_words = comp_words[start:]
 
-        debug("\nLINE: '{l}'\nPREQUOTE: '{pq}'\nPREFIX: '{p}'".format(l=comp_line, pq=cword_prequote, p=cword_prefix),
-              "\nSUFFIX: '{s}'".format(s=cword_suffix),
+        if cword_prefix and cword_prefix[0] in self._parser.prefix_chars and "=" in cword_prefix:
+            # Special case for when the current word is "--optional=PARTIAL_VALUE". Give the optional to the parser.
+            comp_words.append(cword_prefix.split("=", 1)[0])
+
+        debug("\nLINE: {!r}".format(comp_line),
+              "\nPOINT: {!r}".format(comp_point),
+              "\nPREQUOTE: {!r}".format(cword_prequote),
+              "\nPREFIX: {!r}".format(cword_prefix),
+              "\nSUFFIX: {!r}".format(cword_suffix),
               "\nWORDS:", comp_words)
 
         completions = self._get_completions(comp_words, cword_prefix, cword_prequote, last_wordbreak_pos)
@@ -243,7 +252,8 @@ class CompletionFinder(object):
 
         self.completing = False
 
-        # key: complete word, value: description.
+        if "--" in comp_words:
+            self.always_complete_options = False
 
         completions = self.collect_completions(active_parsers, parsed_args, cword_prefix, debug)
         completions = self.filter_completions(completions)
@@ -340,7 +350,8 @@ class CompletionFinder(object):
 
     def _get_option_completions(self, parser, cword_prefix):
         self._display_completions.update(
-            [[" ".join(ensure_str(x) for x in action.option_strings if ensure_str(x).startswith(cword_prefix)), action.help]  # noqa
+            [[" ".join(ensure_str(x) for x in action.option_strings
+                       if ensure_str(x).startswith(cword_prefix)), action.help]
              for action in parser._actions
              if action.option_strings])
 
@@ -371,6 +382,7 @@ class CompletionFinder(object):
         debug("Active actions (L={l}): {a}".format(l=len(parser.active_actions), a=parser.active_actions))
 
         isoptional = cword_prefix and cword_prefix[0] in parser.prefix_chars
+        optional_prefix = ""
         greedy_actions = [x for x in parser.active_actions if action_is_greedy(x, isoptional)]
         if greedy_actions:
             assert len(greedy_actions) == 1, "expect at most 1 greedy action"
@@ -380,8 +392,14 @@ class CompletionFinder(object):
             self._display_completions = {}
             completions = []
         elif isoptional:
-            # Only run completers if current word does not start with - (is not an optional)
-            return completions
+            if "=" in cword_prefix:
+                # Special case for when the current word is "--optional=PARTIAL_VALUE".
+                # The completer runs on PARTIAL_VALUE. The prefix is added back to the completions
+                # (and chopped back off later in quote_completions() by the COMP_WORDBREAKS logic).
+                optional_prefix, _, cword_prefix = cword_prefix.partition("=")
+            else:
+                # Only run completers if current word does not start with - (is not an optional)
+                return completions
 
         complete_remaining_positionals = False
         # Use the single greedy action (if there is one) or all active actions.
@@ -432,6 +450,8 @@ class CompletionFinder(object):
                         if self.validator(next_completion, cword_prefix):
                             self._display_completions.update({next_completion: ""})
                             completions.append(next_completion)
+                if optional_prefix:
+                    completions = [optional_prefix + "=" + completion for completion in completions]
                 debug("Completions:", completions)
         return completions
 
@@ -531,8 +551,8 @@ class CompletionFinder(object):
         elif cword_prequote == '"':
             special_chars += '"`$!'
 
-        if os.environ.get("_ARGCOMPLETE_SHELL") == "tcsh":
-            # tcsh escapes special characters itself.
+        if os.environ.get("_ARGCOMPLETE_SHELL") in ("tcsh", "fish"):
+            # tcsh and fish escapes special characters itself.
             special_chars = ""
         elif cword_prequote == "'":
             # Nothing can be escaped in single quotes, so we need to close
